"""Test general pseudo-element cases."""
from .. import util


class TestPseudoElement(util.TestCase):
    """Test pseudo-elements."""

    def test_pseudo_element(self):
        """Test that pseudo elements always fail because they are not supported."""

        self.assert_raises('::first-line', NotImplementedError)



"""Test checked selectors."""
from .. import util


class TestChecked(util.TestCase):
    """Test checked selectors."""

    def test_checked(self):
        """Test checked."""

        markup = """
        <body>
        <div>
          <input type="radio" name="my-input" id="yes" checked>
          <label for="yes">Yes</label>

          <input type="radio" name="my-input" id="no">
          <label for="no">No</label>
        </div>

        <div>
          <input type="checkbox" name="my-checkbox" id="opt-in" checked>
          <label for="opt-in">Check me!</label>
        </div>

        <select name="my-select" id="fruit">
          <option id="1" value="opt1">Apples</option>
          <option id="2" value="opt2" selected>Grapes</option>
          <option id="3" value="opt3">Pears</option>
        </select>
        </body>
        """

        self.assert_selector(
            markup,
            ":checked",
            ['yes', 'opt-in', '2'],
            flags=util.HTML
        )



"""Test host context selectors."""
from .. import util


class TestHostContext(util.TestCase):
    """Test host context selectors."""

    def test_host_context(self):
        """Test host context (not supported)."""

        markup = """<h1>header</h1><div><p>some text</p></div>"""

        self.assert_selector(
            markup,
            ":host-context(h1, h2)",
            [],
            flags=util.HTML
        )



"""Test root selectors."""
from .. import util
import soupsieve as sv


class TestRoot(util.TestCase):
    """Test root selectors."""

    MARKUP = """
    <html id="root">
    <head>
    </head>
    <body>
    <div id="div">
    <p id="0" class="somewordshere">Some text <span id="1"> in a paragraph</span>.</p>
    <a id="2" href="http://google.com">Link</a>
    <span id="3" class="herewords">Direct child</span>
    <pre id="pre" class="wordshere">
    <span id="4">Child 1</span>
    <span id="5">Child 2</span>
    <span id="6">Child 3</span>
    </pre>
    </div>
    </body>
    </html>
    """

    MARKUP_IFRAME = """
    <html id="root">
    <head>
    </head>
    <body>
    <div id="div">
    </div>
    <iframe src="https://something.com">
    <html id="root2">
    <head>
    </head>
    <body>
    <div id="div2">
    </div>
    </body>
    </html>
    </iframe>
    <div id="other-div"></div>
    </body>
    </html>
    """

    def test_root(self):
        """Test root."""

        # Root in HTML is `<html>`
        self.assert_selector(
            self.MARKUP,
            ":root",
            ["root"],
            flags=util.HTML
        )

    def test_root_iframe(self):
        """Test root."""

        # Root in HTML is `<html>`
        self.assert_selector(
            self.MARKUP_IFRAME,
            ":root",
            ["root", "root2"],
            flags=util.PYHTML
        )

    def test_root_complex(self):
        """Test root within a complex selector."""

        self.assert_selector(
            self.MARKUP,
            ":root > body > div",
            ["div"],
            flags=util.HTML
        )

    def test_no_iframe(self):
        """Test that we don't count `iframe` as root."""

        self.assert_selector(
            self.MARKUP_IFRAME,
            ":root div",
            ["div", "div2", "other-div"],
            flags=util.PYHTML
        )

        self.assert_selector(
            self.MARKUP_IFRAME,
            ":root > body > div",
            ["div", "div2", "other-div"],
            flags=util.PYHTML
        )

    def test_iframe(self):
        """
        Test that we only count `iframe` as root since the scoped element is the root.

        Not all the parsers treat `iframe` content the same. `html5lib` for instance
        will escape the content in the `iframe`, so we are just going to test the builtin
        Python parser.
        """

        soup = self.soup(self.MARKUP_IFRAME, 'html.parser')

        ids = []
        for el in sv.select(':root div', soup.iframe.html):
            ids.append(el['id'])
        self.assertEqual(sorted(ids), sorted(['div2']))

        ids = []
        for el in sv.select(':root > body > div', soup.iframe.html):
            ids.append(el['id'])
        self.assertEqual(sorted(ids), sorted(['div2']))

    def test_no_root_double_tag(self):
        """Test when there is no root due to double root tags."""

        markup = """
        <div id="1"></div>
        <div id="2"></div>
        """

        soup = self.soup(markup, 'html.parser')
        self.assertEqual(soup.select(':root'), [])

    def test_no_root_text(self):
        """Test when there is no root due to HTML text."""

        markup = """
        text
        <div id="1"></div>
        """

        soup = self.soup(markup, 'html.parser')
        self.assertEqual(soup.select(':root'), [])

    def test_no_root_cdata(self):
        """Test when there is no root due to CDATA and tag."""

        markup = """
        <![CDATA[test]]>
        <div id="1"></div>
        """

        soup = self.soup(markup, 'html.parser')
        self.assertEqual(soup.select(':root'), [])

    def test_root_whitespace(self):
        """Test when there is root and white space."""

        markup = """

        <div id="1"></div>
        """

        ids = []
        soup = self.soup(markup, 'html.parser')
        for el in soup.select(':root'):
            ids.append(el['id'])
        self.assertEqual(sorted(ids), sorted(['1']))

    def test_root_preprocess(self):
        """Test when there is root and pre-processing statement."""

        markup = """
        <?php ?>
        <div id="1"></div>
        """

        ids = []
        soup = self.soup(markup, 'html.parser')
        for el in soup.select(':root'):
            ids.append(el['id'])
        self.assertEqual(sorted(ids), sorted(['1']))

    def test_root_doctype(self):
        """Test when there is root and doc type."""

        markup = """
        <!DOCTYPE HTML PUBLIC "-//W3C//DTD HTML 4.01//EN"
        "http://www.w3.org/TR/html4/strict.dtd">
        <div id="1"></div>
        """

        ids = []
        soup = self.soup(markup, 'html.parser')
        for el in soup.select(':root'):
            ids.append(el['id'])
        self.assertEqual(sorted(ids), sorted(['1']))



"""Test language selectors."""
from .. import util


class TestLang(util.TestCase):
    """Test language selectors."""

    MARKUP = """
    <div lang="de-DE">
        <p id="1"></p>
    </div>
    <div lang="de-DE-1996">
        <p id="2"></p>
    </div>
    <div lang="de-Latn-DE">
        <p id="3"></p>
    </div>
    <div lang="de-Latf-DE">
        <p id="4"></p>
    </div>
    <div lang="de-Latn-DE-1996">
        <p id="5"></p>
    </div>
    <p id="6" lang="de-DE"></p>
    <div lang="a-DE">
        <p id="7"></p>
    </div>
    <!-- Singletons don't match implicit wildcards (* not at start are ignored and treated as implicit) -->
    <div lang="a-a-DE">
        <p id="8"></p>
    </div>
    <div lang="en-a-DE">
        <p id="9"></p>
    </div>
    """

    def test_lang(self):
        """Test language and that it uses implicit wildcard."""

        # Implicit wild
        self.assert_selector(
            self.MARKUP,
            "p:lang(de-DE)",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.HTML
        )

    def test_lang_missing_range(self):
        """Test language range with a missing range."""

        # Implicit wild
        self.assert_selector(
            self.MARKUP,
            "p:lang(de--DE)",
            [],
            flags=util.HTML
        )

    def test_explicit_wildcard(self):
        """Test language with explicit wildcard (same as implicit)."""

        # Explicit wild
        self.assert_selector(
            self.MARKUP,
            "p:lang(de-\\*-DE)",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.HTML
        )

    def test_only_wildcard(self):
        """Test language with only a wildcard."""

        self.assert_selector(
            self.MARKUP,
            "p:lang('*')",
            ['1', '2', '3', '4', '5', '6', '7', '8', '9'],
            flags=util.HTML
        )

    def test_wildcard_start_no_match(self):
        """Test language with a wildcard at start, but it matches nothing."""

        self.assert_selector(
            self.MARKUP,
            "p:lang('*-de-DE')",
            [],
            flags=util.HTML
        )

    def test_wildcard_start_collapse(self):
        """Test that language with multiple wildcard patterns at start collapse."""

        self.assert_selector(
            self.MARKUP,
            "p:lang('*-*-*-DE')",
            ['1', '2', '3', '4', '5', '6', '7'],
            flags=util.HTML
        )

    def test_wildcard_at_start_escaped(self):
        """
        Test language with wildcard at start (escaped).

        Wildcard in the middle is same as implicit, but at the start, it has specific meaning.
        """

        self.assert_selector(
            self.MARKUP,
            "p:lang(\\*-DE)",
            ['1', '2', '3', '4', '5', '6', '7'],
            flags=util.HTML
        )

    def test_language_quoted(self):
        """Test language (quoted)."""

        # Normal quoted
        self.assert_selector(
            self.MARKUP,
            "p:lang('de-DE')",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.HTML
        )

    def test_language_quoted_with_escaped_newline(self):
        """Test language (quoted) with escaped new line."""

        # Normal quoted
        self.assert_selector(
            self.MARKUP,
            "p:lang('de-\\\nDE')",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.HTML
        )

    def test_wildcard_at_start_quoted(self):
        """Test language with wildcard at start (quoted)."""

        # First wild quoted
        self.assert_selector(
            self.MARKUP,
            "p:lang('*-DE')",
            ['1', '2', '3', '4', '5', '6', '7'],
            flags=util.HTML
        )

    def test_avoid_implicit_language(self):
        """Test that we can narrow language selection to elements that match and explicitly state language."""

        # Target element with language and language attribute
        self.assert_selector(
            self.MARKUP,
            "p[lang]:lang(de-DE)",
            ['6'],
            flags=util.HTML
        )

    def test_language_list(self):
        """Test language list."""

        # Multiple languages
        markup = """
        <div lang="de-DE">
            <p id="1"></p>
        </div>
        <div lang="en">
            <p id="2"></p>
        </div>
        <div lang="de-Latn-DE">
            <p id="3"></p>
        </div>
        <div lang="de-Latf-DE">
            <p id="4"></p>
        </div>
        <div lang="en-US">
            <p id="5"></p>
        </div>
        <p id="6" lang="de-DE"></p>
        """

        self.assert_selector(
            markup,
            "p:lang(de-DE, '*-US')",
            ['1', '3', '4', '5', '6'],
            flags=util.HTML
        )

    def test_undetermined_language(self):
        """Test undetermined language."""

        markup = """
        <div>
            <p id="1"></p>
        </div>
        """

        self.assert_selector(
            markup,
            "p:lang(en)",
            [],
            flags=util.HTML
        )

    def test_language_in_header(self):
        """Test that we can find language in header."""

        markup = """
        <!DOCTYPE html>
        <html>
        <head>
        <meta http-equiv="content-language" content="en-US">
        </head>
        <body>
        <div>
            <p id="1"></p>
        </div>
        <div>
            <p id="2"></p>
        </div>
        </body>
        """

        self.assert_selector(
            markup,
            "p:lang('*-US')",
            ['1', '2'],
            flags=util.HTML
        )

    def test_xml_style_language_in_html5(self):
        """Test XML style language when out of HTML5 namespace."""

        markup = """
        <math xml:lang="en">
            <mtext id="1"></mtext>
        </math>
        <div xml:lang="en">
            <mtext id="2"></mtext>
        </div>
        """

        self.assert_selector(
            markup,
            "mtext:lang(en)",
            ['1'],
            flags=util.HTML5
        )

    def test_xml_style_language(self):
        """Test XML style language."""

        # XML style language
        markup = """
        <?xml version="1.0" encoding="UTF-8"?>
        <html>
        <head>
        </head>
        <body>
        <div xml:lang="de-DE">
            <p id="1"></p>
        </div>
        <div xml:lang="de-DE-1996">
            <p id="2"></p>
        </div>
        <div xml:lang="de-Latn-DE">
            <p id="3"></p>
        </div>
        <div xml:lang="de-Latf-DE">
            <p id="4"></p>
        </div>
        <div xml:lang="de-Latn-DE-1996">
            <p id="5"></p>
        </div>
        <p id="6" xml:lang="de-DE"></p>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            "p:lang(de-DE)",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.XML
        )

    def test_language_in_xhtml(self):
        """Test language in XHTML."""

        markup = """
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
            "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
        <html lang="en" xmlns="http://www.w3.org/1999/xhtml">
        <head>
        </head>
        <body>
        <div lang="de-DE" xml:lang="de-DE">
            <p id="1"></p>
        </div>
        <div lang="de-DE-1996" xml:lang="de-DE-1996">
            <p id="2"></p>
        </div>
        <div lang="de-Latn-DE" xml:lang="de-Latn-DE">
            <p id="3"></p>
        </div>
        <div lang="de-Latf-DE" xml:lang="de-Latf-DE">
            <p id="4"></p>
        </div>
        <div lang="de-Latn-DE-1996" xml:lang="de-Latn-DE-1996">
            <p id="5"></p>
        </div>
        <p id="6" lang="de-DE" xml:lang="de-DE"></p>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            "p:lang(de-DE)",
            ['1', '2', '3', '4', '5', '6'],
            flags=util.XML
        )

    def test_language_in_xhtml_without_html_style_lang(self):
        """
        Test language in XHTML.

        HTML namespace elements must use HTML style language.
        """

        # XHTML language: `lang`
        markup = """
        <?xml version="1.0" encoding="UTF-8"?>
        <!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.1//EN"
            "http://www.w3.org/TR/xhtml11/DTD/xhtml11.dtd">
        <html lang="en" xmlns="http://www.w3.org/1999/xhtml">
        <head>
        </head>
        <body>
        <div xml:lang="de-DE">
            <p id="1"></p>
        </div>
        <div xml:lang="de-DE-1996">
            <p id="2"></p>
        </div>
        <div xml:lang="de-Latn-DE">
            <p id="3"></p>
        </div>
        <div xml:lang="de-Latf-DE">
            <p id="4"></p>
        </div>
        <div xml:lang="de-Latn-DE-1996">
            <p id="5"></p>
        </div>
        <p id="6" xml:lang="de-DE"></p>
        </body>
        </html>
        """

        self.assert_selector(
            markup,
            "p:lang(de-DE)",
            [],
            flags=util.XHTML
        )



"""Test last of type selectors."""
from .. import util


class TestLastOfType(util.TestCase):
    """Test last of type selectors."""

    def test_last_of_type_at_middle(self):
        """Test last of type that is not the last sibling."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "p:last-of-type",
            ['10'],
            flags=util.HTML
        )

    def test_last_of_type_at_end(self):
        """Test last of type that is the last sibling."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "span:last-of-type",
            ['11'],
            flags=util.HTML
        )

    def test_any_last_of_type(self):
        """Test any last of type."""

        markup = """
        <body>
        <p id="0"></p>
        <p id="1"></p>
        <span id="2"></span>
        <span id="3"></span>
        <span id="4"></span>
        <span id="5"></span>
        <span id="6"></span>
        <p id="7"></p>
        <p id="8"></p>
        <p id="9"></p>
        <p id="10"></p>
        <span id="11"></span>
        </body>
        """

        self.assert_selector(
            markup,
            "body :last-of-type",
            ['10', '11'],
            flags=util.HTML
        )



"""Test subsequent sibling combinator."""
from .. import util


class TestSubsequentSibling(util.TestCase):
    """Test subsequent sibling combinator."""

    def test_subsequent_sibling(self):
        """Test subsequent sibling."""

        self.assert_selector(
            """
            <div>
            <p id="0">Some text <span id="1"> in a paragraph</span>.</p>
            <a id="2" href="http://google.com">Link</a>
            <span id="3">Direct child</span>
            <pre>
            <span id="4">Child 1</span>
            <span id="5">Child 2</span>
            <span id="6">Child 3</span>
            </pre>
            </div>
            """,
            "p ~ span",
            ["3"],
            flags=util.HTML
        )



"""Test contains selectors."""
from .. import util
import warnings
import soupsieve as sv


class TestSoupContains(util.TestCase):
    """Test soup-contains selectors."""

    MARKUP = """
    <body>
    <div id="1">
    Testing
    <span id="2"> that </span>
    contains works.
    </div>
    </body>
    """

    def test_contains(self):
        """Test contains."""

        self.assert_selector(
            self.MARKUP,
            'body span:-soup-contains(that)',
            ['2'],
            flags=util.HTML
        )

    def test_contains_quoted_with_space(self):
        """Test contains quoted with spaces."""

        self.assert_selector(
            self.MARKUP,
            'body span:-soup-contains(" that ")',
            ['2'],
            flags=util.HTML
        )

    def test_contains_quoted_without_space(self):
        """Test contains quoted with spaces."""

        self.assert_selector(
            self.MARKUP,
            'body :-soup-contains( "Testing" )',
            ['1'],
            flags=util.HTML
        )

    def test_contains_quoted_with_escaped_newline(self):
        """Test contains quoted with escaped newline."""

        self.assert_selector(
            self.MARKUP,
            'body :-soup-contains("Test\\\ning")',
            ['1'],
            flags=util.HTML
        )

    def test_contains_quoted_with_escaped_newline_with_carriage_return(self):
        """Test contains quoted with escaped newline with carriage return."""

        self.assert_selector(
            self.MARKUP,
            'body :-soup-contains("Test\\\r\ning")',
            ['1'],
            flags=util.HTML
        )

    def test_contains_list(self):
        """Test contains list."""

        self.assert_selector(
            self.MARKUP,
            'body span:-soup-contains("does not exist", "that")',
            ['2'],
            flags=util.HTML
        )

    def test_contains_multiple(self):
        """Test contains multiple."""

        self.assert_selector(
            self.MARKUP,
            'body span:-soup-contains("th"):-soup-contains("at")',
            ['2'],
            flags=util.HTML
        )

    def test_contains_multiple_not_match(self):
        """Test contains multiple with "not" and with a match."""

        self.assert_selector(
            self.MARKUP,
            'body span:not(:-soup-contains("does not exist")):-soup-contains("that")',
            ['2'],
            flags=util.HTML
        )

    def test_contains_multiple_not_no_match(self):
        """Test contains multiple with "not" and no match."""

        self.assert_selector(
            self.MARKUP,
            'body span:not(:-soup-contains("that")):-soup-contains("that")',
            [],
            flags=util.HTML
        )

    def test_contains_with_descendants(self):
        """Test that contains returns descendants as well as the top level that contain."""

        self.assert_selector(
            self.MARKUP,
            'body :-soup-contains(" that ")',
            ['1', '2'],
            flags=util.HTML
        )

    def test_contains_bad(self):
        """Test contains when it finds no text."""

        self.assert_selector(
            self.MARKUP,
            'body :-soup-contains(bad)',
            [],
            flags=util.HTML
        )

    def test_contains_escapes(self):
        """Test contains with escape characters."""

        markup = """
        <body>
        <div id="1">Testing<span id="2">
        that</span>contains works.</div>
        </body>
        """

        self.assert_selector(
            markup,
            r'body span:-soup-contains("\0a that")',
            ['2'],
            flags=util.HTML
        )

    def test_contains_cdata_html(self):
        """Test contains CDATA in HTML5."""

        markup = """
        <body><div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div></body>
        """

        self.assert_selector(
            markup,
            'body *:-soup-contains("that")',
            ['1'],
            flags=util.HTML
        )

    def test_contains_cdata_xhtml(self):
        """Test contains CDATA in XHTML."""

        markup = """
        <div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div>
        """

        self.assert_selector(
            self.wrap_xhtml(markup),
            'body *:-soup-contains("that")',
            ['1', '2'],
            flags=util.XHTML
        )

    def test_contains_cdata_xml(self):
        """Test contains CDATA in XML."""

        markup = """
        <div id="1">Testing that <span id="2"><![CDATA[that]]></span>contains works.</div>
        """

        self.assert_selector(
            markup,
            '*:-soup-contains("that")',
            ['1', '2'],
            flags=util.XML
        )

    def test_contains_iframe(self):
        """Test contains with `iframe`."""

        markup = """
        <div id="1">
        <p>Testing text</p>
        <iframe>
        <html><body>
        <span id="2">iframe</span>
        </body></html>
        </iframe>
        </div>
        """

        self.assert_selector(
            markup,
            'div:-soup-contains("iframe")',
            [],
            flags=util.PYHTML
        )

        self.assert_selector(
            markup,
            'div:-soup-contains("text")',
            ['1'],
            flags=util.PYHTML
        )

        self.assert_selector(
            markup,
            'span:-soup-contains("iframe")',
            ['2'],
            flags=util.PYHTML
        )

    def test_contains_iframe_xml(self):
        """Test contains with `iframe` which shouldn't matter in XML."""

        markup = """
        <div id="1">
        <p>Testing text</p>
        <iframe>
        <html><body>
        <span id="2">iframe</span>
        </body></html>
        </iframe>
        </div>
        """

        self.assert_selector(
            markup,
            'div:-soup-contains("iframe")',
            ['1'],
            flags=util.XML
        )

        self.assert_selector(
            markup,
            'div:-soup-contains("text")',
            ['1'],
            flags=util.XML
        )

        self.assert_selector(
            markup,
            'span:-soup-contains("iframe")',
            ['2'],
            flags=util.XML
        )

    def test_contains_warn(self):
        """Test old alias raises a warning."""

        sv.purge()

        with warnings.catch_warnings(record=True) as w:
            # Cause all warnings to always be triggered.
            warnings.simplefilter("always")
            # Trigger a warning.
            self.assert_selector(
                self.MARKUP,
                'body span:contains(that)',
                ['2'],
                flags=util.HTML
            )
            # Verify some things
            self.assertTrue(len(w) == 1)
            self.assertTrue(issubclass(w[-1].category, FutureWarning))
